----------------------------------------------------------------------------------
-- BASIC_RS232_INTERFACE.VHD 
-- 
-- Steven Wasson
-- ECE 443 - Hardware Design with VHDL
-- Assignment 6
-- November 29, 2007
--
-- This VHDL module implements a simple RS232 UART.
----------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;
use ieee.math_real.all;

entity BASIC_RS232_INTERFACE is
  generic
  (
    INPUT_CLOCK_FREQUENCY : integer := 1000000;
    BIT_RATE              : integer := 9600
  );
  port
  (
    RESET       : in  std_logic;
    CLOCK       : in  std_logic;

    RTS         : in  std_logic;  -- Request to Send
    DATA_OUT    : in  std_logic_vector(7 downto 0);
    TX_BUSY     : out std_logic;

    RX_INT      : out std_logic;  -- Receiver Interrupt
    FRAME_ERROR : out std_logic;
    RX_ACK      : in  std_logic;
    DATA_IN     : out std_logic_vector(7 downto 0);

    TXD         : out std_logic;
    RXD         : in  std_logic
  );
end BASIC_RS232_INTERFACE;

architecture Behavioral of BASIC_RS232_INTERFACE is

-- Determine the minimum number of bits for the transmitter and receiver timers
constant  TIMER_SIZE  : integer
          := integer(ceil(log2(real(INPUT_CLOCK_FREQUENCY)/real(BIT_RATE))));

-- Determine the bit width in number of CLOCK periods
constant  BIT_WIDTH   : integer
          := integer(round(real(INPUT_CLOCK_FREQUENCY)/real(BIT_RATE)));

signal  shift_out_register      : std_logic_vector(9 downto 0);
signal  tx_timer                : unsigned((TIMER_SIZE-1) downto 0);
signal  tx_bit_counter          : unsigned(3 downto 0);

-- State machine for controlling the transmission of data
type    BASIC_RS232_TRANSMITTER_SM is (WAIT_FOR_RTS, TRANSMIT_DATA);
signal  BASIC_RS232_TRANSMITTER : BASIC_RS232_TRANSMITTER_SM;

signal  shift_in_register       : std_logic_vector(9 downto 0);
signal  rx_timer                : unsigned((TIMER_SIZE-1) downto 0);
signal  rx_bit_counter          : unsigned(3 downto 0);
signal  internal_rx_int         : std_logic;
signal  internal_frame_error    : std_logic;

-- State machine for controlling the reception of data
type    BASIC_RS232_RECEIVER_SM is
        (WAIT_FOR_START_BIT, RECEIVE_DATA, VALIDATE_FRAME);
signal  BASIC_RS232_RECEIVER  : BASIC_RS232_RECEIVER_SM;

begin

  -- Process for transmitting data
  BASIC_RS232_TRANSMITTER_PROCESS:
  process(CLOCK) is
  begin
    if rising_edge(CLOCK) then
      if(RESET = '0') then
        tx_timer <= (others => '0');
        tx_bit_counter <= (others => '0');
        shift_out_register <= (others => '1');
        BASIC_RS232_TRANSMITTER <= WAIT_FOR_RTS;
      else
      
        case BASIC_RS232_TRANSMITTER is
        
          when WAIT_FOR_RTS =>
            if(RTS = '1') then
              BASIC_RS232_TRANSMITTER <= TRANSMIT_DATA;
              tx_timer <= to_unsigned(BIT_WIDTH-1,TIMER_SIZE);
              tx_bit_counter <= (others => '0');
              -- Initialize shift_out_register
              shift_out_register <= '1' & DATA_OUT & '0';
            else
              BASIC_RS232_TRANSMITTER <= WAIT_FOR_RTS;
              shift_out_register <= (others => '1');
            end if;
          
          -- Transmit each bit of data
          -- Hold each bit for a time of BIT_WIDTH
          when TRANSMIT_DATA =>
            if(tx_timer = 0) then
              -- Transmit data until last bit has been shifted out
              if(tx_bit_counter = 9) then 
                BASIC_RS232_TRANSMITTER <= WAIT_FOR_RTS;
                shift_out_register <= (others => '1');
              else
                BASIC_RS232_TRANSMITTER <= TRANSMIT_DATA;
              end if;
              tx_timer <= to_unsigned(BIT_WIDTH-1,TIMER_SIZE);
              tx_bit_counter <= tx_bit_counter + 1;
              shift_out_register <= '1' & shift_out_register(9 downto 1);
            else
              tx_timer <= tx_timer - 1;
            end if;

        end case;
        
      end if;
    end if;
  end process;

  TXD <= shift_out_register(0);
  TX_BUSY <= '1' when (BASIC_RS232_TRANSMITTER = TRANSMIT_DATA) else '0';

  -- Process for receiving data
  BASIC_RS232_RECEIVER_PROCESS:
  process(CLOCK) is
  begin
    if rising_edge(CLOCK) then
      if(RESET = '0') then
        rx_timer <= (others => '0');
        rx_bit_counter <= (others => '0');
        shift_in_register <= (others => '0');
        internal_frame_error <= '0';
        internal_rx_int <= '0';
        BASIC_RS232_RECEIVER <= WAIT_FOR_START_BIT;
      else
      
        case BASIC_RS232_RECEIVER is
        
          when WAIT_FOR_START_BIT =>
            if(RXD = '0') then
              BASIC_RS232_RECEIVER <= RECEIVE_DATA;
              -- Initialize the receiver timer such that it
              -- times out in the middle of the start bit
              rx_timer <= to_unsigned((BIT_WIDTH/2)-1,TIMER_SIZE);
              rx_bit_counter <= (others => '0');
              shift_in_register <= (others => '0');
              internal_frame_error <= '0';
              internal_rx_int <= '0';
            else
              BASIC_RS232_RECEIVER <= WAIT_FOR_START_BIT;
            end if;

          -- Receive the bits of data
          -- Clock in each bit after a time of BIT_WIDTH
          when RECEIVE_DATA =>
            if(rx_timer = 0) then
              if(rx_bit_counter = 9) then
                BASIC_RS232_RECEIVER <= VALIDATE_FRAME;
              else
                BASIC_RS232_RECEIVER <= RECEIVE_DATA;
              end if;
              rx_timer <= to_unsigned(BIT_WIDTH-1,TIMER_SIZE);
              rx_bit_counter <= rx_bit_counter + 1;
              shift_in_register <= RXD & shift_in_register(9 downto 1);
            else
              rx_timer <= rx_timer - 1;
            end if;

          -- Validate the input frame by checking the start and stop bits
          when VALIDATE_FRAME =>
            if((shift_in_register(0) = '0') and (shift_in_register(9) = '1')) then
              internal_rx_int <= '1';
            else
              internal_frame_error <= '1';
            end if;
            BASIC_RS232_RECEIVER <= WAIT_FOR_START_BIT;

        end case;

        -- Clear RX_INT when RX_ACK is high
        if((internal_rx_int = '1') and (RX_ACK = '1')) then
          internal_rx_int <= '0';
        end if;

        -- Clear FRAME_ERROR when RX_ACK is high
        if((internal_frame_error = '1') and (RX_ACK = '1')) then
          internal_frame_error <= '0';
        end if;
        
      end if;
    end if;
  end process;
  
  DATA_IN <= shift_in_register(8 downto 1);
  RX_INT <= internal_rx_int;
  FRAME_ERROR <= internal_frame_error;

end Behavioral;
