library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity sound_recorder is
  port
  (
    -- Signals used to produce asynchronous reset to CODEC
    SYS_CLK               : in  std_logic;  -- 100 MHz
    RESET_IN              : in  std_logic;
    
    -- Record and Play Switches and LEDs
    BUTTON                : in  std_logic_vector(0 to 1);
    RECORD_LED            : out std_logic;
    PLAY_LED              : out std_logic;

    -- CODEC Signals
    BIT_CLK               : in  std_logic;  -- 12.288 MHz
    RESET_OUT             : out std_logic;
    SYNC                  : out std_logic;
    SDATA_OUT             : out std_logic;
    SDATA_IN              : in  std_logic;
    
    -- RS232 Signals
    TXD         : out std_logic;
    RXD         : in  std_logic
  );
end sound_recorder;

architecture Behavioral of sound_recorder is

-- Define SWITCH_DEBOUNCER component
component SWITCH_DEBOUNCER
  generic
  (
    WIDTH                   : integer;
    DELAY                   : integer
  );

  port
  (
    RESET                   : in std_logic;
    CLOCK                   : in std_logic;
    INPUT                   : in std_logic_vector((WIDTH-1) downto 0);
    OUTPUT                  : out std_logic_vector((WIDTH-1) downto 0);
    NEW_OUTPUT              : out std_logic;
    ACKNOWLEDGE             : in std_logic
  );
end component;

-- Define BASIC_AC9Y_INTERFACE component
component BASIC_AC97_INTERFACE
  generic
  (
    SAMPLING_RATE                 : integer
  );
  port
  (
    -- Signals used to produce asynchronous reset to CODEC
    SYS_CLK                       : in  std_logic;  -- 100 MHz
    RESET_IN                      : in  std_logic;

    -- CODEC Signals
    BIT_CLK                       : in  std_logic;  -- 12.288 MHz
    RESET_OUT                     : out std_logic;
    SYNC                          : out std_logic;
    SDATA_OUT                     : out std_logic;
    SDATA_IN                      : in  std_logic;

    -- Interface Signals
    BUSY                          : out std_logic;
    READY_FOR_AUDIO_TRANSMISSION  : out std_logic;
    VALID_AUDIO_INPUT             : out std_logic;
    LEFT_AUDIO_INPUT              : out std_logic_vector(17 downto 0);
    RIGHT_AUDIO_INPUT             : out std_logic_vector(17 downto 0);
    VALID_AUDIO_OUTPUT            : in  std_logic;
    LEFT_AUDIO_OUTPUT             : in  std_logic_vector(17 downto 0);
    RIGHT_AUDIO_OUTPUT            : in  std_logic_vector(17 downto 0)    
  );
end component;

-- Define AUDIO_MEMORY BLOCK
component AUDIO_MEMORY
	port
  (
    addr  : in  std_logic_VECTOR(16 downto 0);
    clk   : in  std_logic;
    din   : in  std_logic_VECTOR(17 downto 0);
    dout  : out std_logic_VECTOR(17 downto 0);
    we    : in  std_logic
  );
end component;

-- Define CONSTANT_CLOCK_DIVIDER component
component CONSTANT_CLOCK_DIVIDER
  generic
  (
    DIVIDER       : integer
  );
  port
  (
    CLOCK_IN      : in  std_logic;
    CLOCK_OUT     : out std_logic
  );
end component;

-- Define ASCII_ROM component
component ASCII_ROM
	port
  (
    clka:   in  std_logic;
    addra:  in  std_logic_VECTOR(13 downto 0);
    douta:  out std_logic_VECTOR(6 downto 0)
  );
end component;

-- Define BASIC_RS232_INTERFACE component
component BASIC_RS232_INTERFACE
  generic
  (
    INPUT_CLOCK_FREQUENCY : integer;
    BIT_RATE              : integer
  );
  port
  (
    RESET       : in  std_logic;
    CLOCK       : in  std_logic;

    RTS         : in  std_logic;  -- Request to Send
    DATA_OUT    : in  std_logic_vector(7 downto 0);
    TX_BUSY     : out std_logic;

    RX_INT      : out std_logic;  -- Receiver Interrupt
    FRAME_ERROR : out std_logic;
    RX_ACK      : in  std_logic;
    DATA_IN     : out std_logic_vector(7 downto 0);

    TXD         : out std_logic;
    RXD         : in  std_logic
  );
end component BASIC_RS232_INTERFACE;

-- Record/Play button signals
signal    DEBOUNCED_BUTTON              : std_logic_vector(0 to 1);
signal    NEW_BUTTON_STATE              : std_logic;

-- Basic RS232 Interface Signals
signal    RS232_CLOCK                   : std_logic;
signal    RTS                           : std_logic;
signal    ASCII_DATA_OUT                : std_logic_vector(7 downto 0);
signal    TX_BUSY                       : std_logic;
signal    RX_INT                        : std_logic;
signal    FRAME_ERROR                   : std_logic;
signal    RX_ACK                        : std_logic;
signal    ASCII_DATA_IN                 : std_logic_vector(7 downto 0);

-- ASCII Memory Signals
signal    ASCII_MEMORY_POINTER          : std_logic_vector(13 downto 0);
signal    LAST_ASCII_MEMORY_ADDRESS     : std_logic_vector(13 downto 0);
constant  BUGS_BUNNY_1    : std_logic_vector(13 downto 0) := "00000000000000";
constant  PEPE_LEPEW      : std_logic_vector(13 downto 0) := "00100101110010";
constant  TWEETY_BIRD     : std_logic_vector(13 downto 0) := "01000000101000";
constant  BUGS_BUNNY_2    : std_logic_vector(13 downto 0) := "01100010010110";
constant  TASMANIAN_DEVIL : std_logic_vector(13 downto 0) := "01111011100011";
constant  DAFFY_DUCK      : std_logic_vector(13 downto 0) := "10001101110000";
constant  ASCII_ROM_SIZE  : std_logic_vector(13 downto 0) := "10100100000100";

-- Basic AC97 Interface Signals
signal    BUSY                          : std_logic;
signal    READY_FOR_AUDIO_TRANSMISSION  : std_logic;
signal    VALID_AUDIO_INPUT             : std_logic;
signal    MONO_AUDIO_INPUT              : std_logic_vector(17 downto 0);
signal    MONO_AUDIO_OUTPUT             : std_logic_vector(17 downto 0);

-- Audio Memory Signals
signal    AUDIO_MEMORY_POINTER          : std_logic_vector(16 downto 0);
signal    LAST_AUDIO_SAMPLE_ADDRESS     : std_logic_vector(16 downto 0);
signal    LOOPBACK_AUDIO_OUTPUT         : std_logic_vector(17 downto 0);
signal    RECORDED_AUDIO_OUTPUT         : std_logic_vector(17 downto 0);
signal    RECORD_IN_PROGRESS            : std_logic;
signal    PLAY_IN_PROGRESS              : std_logic;
signal    RECORD_ENABLE                 : std_logic;
signal    END_OF_AUDIO                  : std_logic;
signal    END_OF_AUDIO_ACKNOWLEDGEMENT  : std_logic;

-- Finite State Machine for controlling the whole system
type      SOUND_RECORDER_SM is 
          (WAIT_FOR_INPUT, ACKNOWLEDGE_TERMINAL_INPUT,
           WAIT_FOR_INACTIVE_BUTTONS, ACKNOWLEDGE_END_OF_AUDIO, EXIT_PROGRAM);
signal    SOUND_RECORDER_CURRENT_TASK   : SOUND_RECORDER_SM;

-- Finite State Machine for outputting ASCII art to terminal interface
type      TERMINAL_OUTPUT_SM is 
          (WAIT_FOR_COMMAND, WAIT_FOR_TX_BUSY_TOGO_HIGH, WAIT_FOR_TX_BUSY_TOGO_LOW,
           TRANSMIT_CHARACTERS, INCREMENT_ADDRESS);
signal    TERMINAL_OUTPUT_CURRENT_TASK  : TERMINAL_OUTPUT_SM;

-- Signal to control the TERMINAL_OUTPUT state machine
type      TERMINAL_COMMAND_TYPE is
          (IDLE, OUTPUT_BUGS_BUNNY_1, OUTPUT_PEPE_LEPEW, OUTPUT_TWEETY_BIRD,
           OUTPUT_BUGS_BUNNY_2, OUTPUT_TASMANIAN_DEVIL, OUTPUT_DAFFY_DUCK);
signal    TERMINAL_COMMAND              : TERMINAL_COMMAND_TYPE;

-- Finite State Machine for controlling recording and playback activity
type      RECORD_PLAY_SM is
          (WAIT_FOR_COMMAND, WAIT_FOR_BUSY_TOGO_HIGH, WAIT_FOR_BUSY_TOGO_LOW,
           CHECK_CODEC_STATUS, RECORD_PLAY_SAMPLE);
signal    RECORD_PLAY_CURRENT_TASK      : RECORD_PLAY_SM;

-- Signal to control the RECORD_PLAY state machine
type      RECORD_PLAY_COMMAND_TYPE is
          (RECORD_COMMAND, PLAY_COMMAND, STOP_COMMAND);
signal    RECORD_PLAY_COMMAND           : RECORD_PLAY_COMMAND_TYPE;

begin

  -- Debounce the record/play buttons
  DEBOUNCE_BUTTONS: SWITCH_DEBOUNCER
  generic map
  (
    WIDTH                   => 2,
    DELAY                   => 16
  )
  port map
  (
    RESET                   => '0',
    CLOCK                   => BIT_CLK,
    INPUT                   => BUTTON,
    OUTPUT                  => DEBOUNCED_BUTTON,
    NEW_OUTPUT              => NEW_BUTTON_STATE,
    ACKNOWLEDGE             => NEW_BUTTON_STATE
  );
  
  GENERATE_RS232_CLOCK:
  CONSTANT_CLOCK_DIVIDER
  generic map
  (
    DIVIDER       => 6
  )
  port map
  (
    CLOCK_IN      => BIT_CLK,     -- 12.288 MHz
    CLOCK_OUT     => RS232_CLOCK  -- 1.024  MHz
  );
  
  RS232_INSTANCE:
  BASIC_RS232_INTERFACE
  generic map
  (
    INPUT_CLOCK_FREQUENCY => 1024000,
    BIT_RATE              => 115200
  )
  port map
  (
    RESET       => RESET_IN,
    CLOCK       => RS232_CLOCK,

    RTS         => RTS,
    DATA_OUT    => ASCII_DATA_OUT,
    TX_BUSY     => TX_BUSY,

    RX_INT      => RX_INT,
    FRAME_ERROR => FRAME_ERROR,
    RX_ACK      => RX_ACK,
    DATA_IN     => ASCII_DATA_IN,

    TXD         => TXD,
    RXD         => RXD
  );
  
  SOUND_RECORDER_PROCESS:
  process(BIT_CLK) is
  begin
    if(RESET_IN = '0') then
      SOUND_RECORDER_CURRENT_TASK <= WAIT_FOR_INPUT;
      TERMINAL_COMMAND <= IDLE;
      RECORD_PLAY_COMMAND <= STOP_COMMAND;

    elsif rising_edge(BIT_CLK) then
      
      case SOUND_RECORDER_CURRENT_TASK is
        
        when WAIT_FOR_INPUT =>
          TERMINAL_COMMAND <= IDLE;
          if(RX_INT = '1') then
            SOUND_RECORDER_CURRENT_TASK <= ACKNOWLEDGE_TERMINAL_INPUT;
            if(FRAME_ERROR = '0') then
              case ASCII_DATA_IN is
                when X"48" =>
                  TERMINAL_COMMAND <= OUTPUT_BUGS_BUNNY_1;
                when X"52" =>
                  TERMINAL_COMMAND <= OUTPUT_PEPE_LEPEW;
                  RECORD_PLAY_COMMAND <= RECORD_COMMAND;
                when X"50" =>
                  TERMINAL_COMMAND <= OUTPUT_TWEETY_BIRD;
                  RECORD_PLAY_COMMAND <= PLAY_COMMAND;
                when X"53" =>
                  TERMINAL_COMMAND <= OUTPUT_BUGS_BUNNY_2;
                  RECORD_PLAY_COMMAND <= STOP_COMMAND;
                when X"58" =>
                  TERMINAL_COMMAND <= OUTPUT_TASMANIAN_DEVIL;
                when others =>
                  TERMINAL_COMMAND <= OUTPUT_DAFFY_DUCK;
              end case;
            end if;
          elsif(NEW_BUTTON_STATE = '1') then
            SOUND_RECORDER_CURRENT_TASK <= WAIT_FOR_INACTIVE_BUTTONS;
            case DEBOUNCED_BUTTON is
              when "01" =>
                if(RECORD_IN_PROGRESS = '0') then
                  TERMINAL_COMMAND <= OUTPUT_PEPE_LEPEW;
                  RECORD_PLAY_COMMAND <= RECORD_COMMAND;
                else
                  TERMINAL_COMMAND <= OUTPUT_BUGS_BUNNY_2;
                  RECORD_PLAY_COMMAND <= STOP_COMMAND;
                end if;
              when "10" =>
                if(PLAY_IN_PROGRESS = '0') then
                  TERMINAL_COMMAND <= OUTPUT_TWEETY_BIRD;
                  RECORD_PLAY_COMMAND <= PLAY_COMMAND;
                else
                  TERMINAL_COMMAND <= OUTPUT_BUGS_BUNNY_2;
                  RECORD_PLAY_COMMAND <= STOP_COMMAND;
                end if;
              when others =>
                TERMINAL_COMMAND <= OUTPUT_DAFFY_DUCK;
            end case;
          elsif(END_OF_AUDIO = '1') then
            SOUND_RECORDER_CURRENT_TASK <= ACKNOWLEDGE_END_OF_AUDIO;
            TERMINAL_COMMAND <= OUTPUT_BUGS_BUNNY_2;
            RECORD_PLAY_COMMAND <= STOP_COMMAND;
          else
            SOUND_RECORDER_CURRENT_TASK <= WAIT_FOR_INPUT;
          end if;
          
        when ACKNOWLEDGE_TERMINAL_INPUT =>

          if(RX_INT = '0') then
            if(TERMINAL_COMMAND /= OUTPUT_TASMANIAN_DEVIL) then
              SOUND_RECORDER_CURRENT_TASK <= WAIT_FOR_INPUT;
            else
              SOUND_RECORDER_CURRENT_TASK <= EXIT_PROGRAM;
            end if;
          else
            SOUND_RECORDER_CURRENT_TASK <= ACKNOWLEDGE_TERMINAL_INPUT;
          end if;
          
        when WAIT_FOR_INACTIVE_BUTTONS =>

          if(DEBOUNCED_BUTTON = "11") then
            SOUND_RECORDER_CURRENT_TASK <= WAIT_FOR_INPUT;
          else
            SOUND_RECORDER_CURRENT_TASK <= WAIT_FOR_INACTIVE_BUTTONS;
          end if;
        
        when ACKNOWLEDGE_END_OF_AUDIO =>

          if(END_OF_AUDIO = '0') then
            SOUND_RECORDER_CURRENT_TASK <= WAIT_FOR_INPUT;
          else
            SOUND_RECORDER_CURRENT_TASK <= ACKNOWLEDGE_END_OF_AUDIO;
          end if;
        
        when EXIT_PROGRAM =>
          SOUND_RECORDER_CURRENT_TASK <= EXIT_PROGRAM;
          TERMINAL_COMMAND <= IDLE;
          RECORD_PLAY_COMMAND <= STOP_COMMAND;
          
      end case;

    end if;
  end process;
  
  RX_ACK <= '1' when (SOUND_RECORDER_CURRENT_TASK = ACKNOWLEDGE_TERMINAL_INPUT) else '0';

  END_OF_AUDIO_ACKNOWLEDGEMENT <= '1'
    when (SOUND_RECORDER_CURRENT_TASK = ACKNOWLEDGE_END_OF_AUDIO) else '0';
  
  ASCII_ART_INTERFACE:
  ASCII_ROM
	port map
  (
    clka  => BIT_CLK,
    addra => ASCII_MEMORY_POINTER,
    douta => ASCII_DATA_OUT(6 downto 0)
  );
  
  ASCII_DATA_OUT(7) <= '0';
  
  TERMINAL_OUTPUT_PROCESS:
  process(RESET_IN, BIT_CLK) is
  begin
    if(RESET_IN = '0') then
      RTS <= '0';
      ASCII_MEMORY_POINTER <= BUGS_BUNNY_1;
      LAST_ASCII_MEMORY_ADDRESS <= PEPE_LEPEW;
      TERMINAL_OUTPUT_CURRENT_TASK <= TRANSMIT_CHARACTERS;

    elsif rising_edge(BIT_CLK) then
    
      case TERMINAL_OUTPUT_CURRENT_TASK is
      
        when WAIT_FOR_COMMAND =>
          
          if(TERMINAL_COMMAND /= IDLE) then
            
            TERMINAL_OUTPUT_CURRENT_TASK <= TRANSMIT_CHARACTERS;
              
            case TERMINAL_COMMAND is
              when OUTPUT_BUGS_BUNNY_1 =>
                ASCII_MEMORY_POINTER <= BUGS_BUNNY_1;
                LAST_ASCII_MEMORY_ADDRESS <= PEPE_LEPEW;
              when OUTPUT_PEPE_LEPEW =>
                ASCII_MEMORY_POINTER <= PEPE_LEPEW;
                LAST_ASCII_MEMORY_ADDRESS <= TWEETY_BIRD;
              when OUTPUT_TWEETY_BIRD =>
                ASCII_MEMORY_POINTER <= TWEETY_BIRD;
                LAST_ASCII_MEMORY_ADDRESS <= BUGS_BUNNY_2;
              when OUTPUT_BUGS_BUNNY_2 =>
                ASCII_MEMORY_POINTER <= BUGS_BUNNY_2;
                LAST_ASCII_MEMORY_ADDRESS <= TASMANIAN_DEVIL;
              when OUTPUT_TASMANIAN_DEVIL =>
                ASCII_MEMORY_POINTER <= TASMANIAN_DEVIL;
                LAST_ASCII_MEMORY_ADDRESS <= DAFFY_DUCK;
              when OUTPUT_DAFFY_DUCK =>
                ASCII_MEMORY_POINTER <= DAFFY_DUCK;
                LAST_ASCII_MEMORY_ADDRESS <= ASCII_ROM_SIZE;
              when others =>
                ASCII_MEMORY_POINTER <= BUGS_BUNNY_1;
                LAST_ASCII_MEMORY_ADDRESS <= PEPE_LEPEW;
            end case;
            
          else
            
            TERMINAL_OUTPUT_CURRENT_TASK <= WAIT_FOR_COMMAND;
              
          end if;

        when WAIT_FOR_TX_BUSY_TOGO_HIGH =>
          if(TX_BUSY = '1') then
            RTS <= '0';
            TERMINAL_OUTPUT_CURRENT_TASK <= WAIT_FOR_TX_BUSY_TOGO_LOW;
          else
            TERMINAL_OUTPUT_CURRENT_TASK <= WAIT_FOR_TX_BUSY_TOGO_HIGH;
          end if;
        
        when WAIT_FOR_TX_BUSY_TOGO_LOW =>
          if(TX_BUSY = '0') then
            TERMINAL_OUTPUT_CURRENT_TASK <= INCREMENT_ADDRESS;
          else
            TERMINAL_OUTPUT_CURRENT_TASK <= WAIT_FOR_TX_BUSY_TOGO_LOW;
          end if;

        when TRANSMIT_CHARACTERS =>
          if(ASCII_MEMORY_POINTER /= LAST_ASCII_MEMORY_ADDRESS) then
            TERMINAL_OUTPUT_CURRENT_TASK <= WAIT_FOR_TX_BUSY_TOGO_HIGH;
            RTS <= '1';
          else
            TERMINAL_OUTPUT_CURRENT_TASK <= WAIT_FOR_COMMAND;
          end if;
        
        when INCREMENT_ADDRESS =>
          TERMINAL_OUTPUT_CURRENT_TASK <= TRANSMIT_CHARACTERS;
          ASCII_MEMORY_POINTER <= ASCII_MEMORY_POINTER + '1';

      end case;

    end if;
  end process;
  
  BASIC_AC97_INTERFACE_INSTANCE:
  BASIC_AC97_INTERFACE
  generic map
  (
    SAMPLING_RATE                 => 4000 -- Hz
  )
  port map
  (
    -- Signals used to produce asynchronous reset to CODEC
    SYS_CLK                       => SYS_CLK,
    RESET_IN                      => RESET_IN,

    -- CODEC Signals
    BIT_CLK                       => BIT_CLK,
    RESET_OUT                     => RESET_OUT,
    SYNC                          => SYNC,
    SDATA_OUT                     => SDATA_OUT,
    SDATA_IN                      => SDATA_IN,

    -- Interface Signals
    BUSY                          => BUSY,
    READY_FOR_AUDIO_TRANSMISSION  => READY_FOR_AUDIO_TRANSMISSION,----
    VALID_AUDIO_INPUT             => VALID_AUDIO_INPUT,--             }
    LEFT_AUDIO_INPUT              => MONO_AUDIO_INPUT,--              |
    RIGHT_AUDIO_INPUT             => open,--                          }
    VALID_AUDIO_OUTPUT            => READY_FOR_AUDIO_TRANSMISSION,----
    LEFT_AUDIO_OUTPUT             => MONO_AUDIO_OUTPUT,
    RIGHT_AUDIO_OUTPUT            => MONO_AUDIO_OUTPUT
  );
  
  LOOPBACK_AUDIO_PROCESS: process(BIT_CLK) is
  begin
    if rising_edge(BIT_CLK) then
      if((BUSY = '0') and (VALID_AUDIO_INPUT = '1')) then
        LOOPBACK_AUDIO_OUTPUT <= MONO_AUDIO_INPUT;
      end if;
    end if;
  end process;

  AUDIO_MEMORY_INSTANCE:
  AUDIO_MEMORY
  port map
  (
    addr  => AUDIO_MEMORY_POINTER,
    clk   => BIT_CLK,
    din   => MONO_AUDIO_INPUT,
    dout  => RECORDED_AUDIO_OUTPUT,
    we    => RECORD_ENABLE
  );
  
  MONO_AUDIO_OUTPUT <= LOOPBACK_AUDIO_OUTPUT when (RECORD_IN_PROGRESS = '1') else
                       RECORDED_AUDIO_OUTPUT when (PLAY_IN_PROGRESS = '1') else
                       (others => '0');

  RECORD_AND_PLAYBACK: process(RESET_IN, BIT_CLK) is
  begin
    if(RESET_IN = '0') then
      AUDIO_MEMORY_POINTER <= (others => '0');
      RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_COMMAND;
      RECORD_IN_PROGRESS <= '0';
      PLAY_IN_PROGRESS <= '0';
      RECORD_ENABLE <= '0';
    
    elsif rising_edge(BIT_CLK) then
      
      case RECORD_PLAY_CURRENT_TASK is
      
        -- Wait until the RECORD or PLAY command has been issued
        when WAIT_FOR_COMMAND =>
          if((RECORD_PLAY_COMMAND = RECORD_COMMAND) and (END_OF_AUDIO = '0')) then
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_HIGH;
            AUDIO_MEMORY_POINTER <= (others => '0');
            RECORD_IN_PROGRESS <= '1';
          elsif((RECORD_PLAY_COMMAND = PLAY_COMMAND) and (END_OF_AUDIO = '0')) then
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_HIGH;
            AUDIO_MEMORY_POINTER <= (others => '0');
            PLAY_IN_PROGRESS <= '1';
          else
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_COMMAND;
          end if;
          
        -- Wait until BUSY goes high
        when WAIT_FOR_BUSY_TOGO_HIGH =>
          if(BUSY = '1') then
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_LOW;
          elsif(RECORD_PLAY_COMMAND = STOP_COMMAND) then
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_COMMAND;
            if(RECORD_IN_PROGRESS = '1') then
              LAST_AUDIO_SAMPLE_ADDRESS <= AUDIO_MEMORY_POINTER;
            end if;
            RECORD_IN_PROGRESS <= '0';
            PLAY_IN_PROGRESS <= '0';
          else
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_HIGH;
          end if;
        
        -- Wait until BUSY goes low
        when WAIT_FOR_BUSY_TOGO_LOW =>
          if(BUSY = '0') then
            RECORD_PLAY_CURRENT_TASK <= CHECK_CODEC_STATUS;
          elsif(RECORD_PLAY_COMMAND = STOP_COMMAND) then
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_COMMAND;
            if(RECORD_IN_PROGRESS = '1') then
              LAST_AUDIO_SAMPLE_ADDRESS <= AUDIO_MEMORY_POINTER;
            end if;
            RECORD_IN_PROGRESS <= '0';
            PLAY_IN_PROGRESS <= '0';
          else
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_LOW;
          end if;
        
        -- Check to see if the CODEC is ready for data
        when CHECK_CODEC_STATUS =>
          if((RECORD_IN_PROGRESS = '1') and (VALID_AUDIO_INPUT = '1')) then
            RECORD_PLAY_CURRENT_TASK <= RECORD_PLAY_SAMPLE;
            RECORD_ENABLE <= '1';
          elsif((PLAY_IN_PROGRESS = '1') and (READY_FOR_AUDIO_TRANSMISSION = '1')) then
            RECORD_PLAY_CURRENT_TASK <= RECORD_PLAY_SAMPLE;
          else
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_HIGH;
          end if;
        
        -- Record/Play the current audio sample
        when RECORD_PLAY_SAMPLE =>
          if((RECORD_IN_PROGRESS = '1') and (AUDIO_MEMORY_POINTER = X"1FFFF")) then
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_COMMAND;
            LAST_AUDIO_SAMPLE_ADDRESS <= (others => '1');
            RECORD_IN_PROGRESS <= '0';
            END_OF_AUDIO <= '1';
          elsif((PLAY_IN_PROGRESS = '1') and (AUDIO_MEMORY_POINTER = LAST_AUDIO_SAMPLE_ADDRESS)) then
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_COMMAND;
            PLAY_IN_PROGRESS <= '0';
            END_OF_AUDIO <= '1';
          else
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_HIGH;
          end if;
          
          AUDIO_MEMORY_POINTER <= AUDIO_MEMORY_POINTER + '1';
          RECORD_ENABLE <= '0';
      
      end case;
      
      if((END_OF_AUDIO = '1') and (END_OF_AUDIO_ACKNOWLEDGEMENT = '1')) then
        END_OF_AUDIO <= '0';
      end if;
    
    end if;
    
  end process;
  
  RECORD_LED <= not RECORD_IN_PROGRESS;
  PLAY_LED <= not PLAY_IN_PROGRESS;

end Behavioral;
