----------------------------------------------------------------------------------
-- BASIC_AC97_CORE.VHD 
-- 
-- Steven Wasson
-- ECE 443 - Hardware Design with VHDL
-- Assignment 4
-- October 31, 2007
--
-- This VHDL module handles the serial communications to the AC97 CODEC.
----------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity BASIC_AC97_CORE is
  port
  (
    -- CODEC Signals
    BIT_CLK                         : in  std_logic;
    RESET_IN                        : in  std_logic;
    RESET_OUT                       : out std_logic;
    SYNC                            : out std_logic;
    SDATA_OUT                       : out std_logic;
    SDATA_IN                        : in  std_logic;
    BUSY                            : out std_logic;
    
    -- Input Interface Signals
    VALID_CONTROL_REGISTER_ADDRESS  : in  std_logic;
    VALID_CONTROL_REGISTER_DATA     : in  std_logic;
    VALID_AUDIO_OUTPUT              : in  std_logic;
    CONTROL_REGISTER_ADDRESS        : in  std_logic_vector(7 downto 0);
    CONTROL_REGISTER_DATA           : in  std_logic_vector(15 downto 0);
    LEFT_AUDIO_OUTPUT               : in  std_logic_vector(17 downto 0);
    RIGHT_AUDIO_OUTPUT              : in  std_logic_vector(17 downto 0);
    
    -- Output Interface Signals
    CODEC_READY                     : out std_logic;
    VALID_STATUS_ADDRESS            : out std_logic;
    VALID_STATUS_DATA               : out std_logic;
    VALID_AUDIO_INPUT               : out std_logic;
    STATUS_REGISTER_INDEX           : out std_logic_vector(6 downto 0);
    LEFT_AUDIO_REQUEST              : out std_logic;
    RIGHT_AUDIO_REQUEST             : out std_logic;
    STATUS_DATA                     : out std_logic_vector(15 downto 0);
    LEFT_AUDIO_INPUT                : out std_logic_vector(17 downto 0);
    RIGHT_AUDIO_INPUT               : out std_logic_vector(17 downto 0)
  );
end BASIC_AC97_CORE;

architecture Behavioral of BASIC_AC97_CORE is

-- AC97 State Machine Controller operating from BIT_CLK
type    AC97_CONTROLLER_SM is (TRANSMIT_LAST_BIT_OF_FRAME, TRANSMIT_TAG,
                               TRANSMIT_SLOTS);
signal  AC97_CONTROLLER   : AC97_CONTROLLER_SM;

signal  VALID_FRAME       : std_logic;
signal  INTERNAL_BUSY     : std_logic;
signal  INTERNAL_SYNC     : std_logic;
signal  FRAME_DATA        : std_logic_vector(0 to 93);    -- Output
signal  BIT_COUNTER       : std_logic_vector(7 downto 0);
signal  SHIFT_REGISTER    : std_logic_vector(0 to 93);    -- Input

begin

  -- A valid output frame is present when there is a valid control register
  -- address, valid control register data, or valid audio data
  VALID_FRAME <= (VALID_CONTROL_REGISTER_ADDRESS or
                  VALID_CONTROL_REGISTER_DATA or
                  VALID_AUDIO_OUTPUT);

  SIMPLE_AC97_CONTROLLER:  process(RESET_IN, BIT_CLK) is
  begin

    -- Asynchronous reset is necessary because when the reset
    -- is asserted to the CODEC the BIT_CLK is disabled
    if(RESET_IN = '0') then
      AC97_CONTROLLER <= TRANSMIT_LAST_BIT_OF_FRAME;
      INTERNAL_BUSY <= '1';
      INTERNAL_SYNC <= '0';
      BIT_COUNTER <= X"FF";
      FRAME_DATA <= (others => '0');
    
    else

      if rising_edge(BIT_CLK) then

        case AC97_CONTROLLER is

          when TRANSMIT_LAST_BIT_OF_FRAME =>  -- TRANSMIT_LAST_BIT_OF_FRAME state
            if(VALID_FRAME = '1') then
              -- Construct Tag
              FRAME_DATA(0) <= '1';                                  -- Valid Frame
              FRAME_DATA(1) <= VALID_CONTROL_REGISTER_ADDRESS;
              FRAME_DATA(2) <= VALID_CONTROL_REGISTER_DATA;
              FRAME_DATA(3) <= VALID_AUDIO_OUTPUT;
              FRAME_DATA(4) <= VALID_AUDIO_OUTPUT;
              FRAME_DATA(5 to 15) <= (others => '0');
              FRAME_DATA(16 to 35) <= CONTROL_REGISTER_ADDRESS & X"000";  -- Slot 1
              FRAME_DATA(36 to 55) <= CONTROL_REGISTER_DATA & X"0";       -- Slot 2
              FRAME_DATA(56 to 75) <= LEFT_AUDIO_OUTPUT & "00";           -- Slot 3
              FRAME_DATA(76 to 93) <= RIGHT_AUDIO_OUTPUT;                 -- Slot 4
            else
              FRAME_DATA <= (others => '0');
            end if;
            AC97_CONTROLLER <= TRANSMIT_TAG;
            INTERNAL_BUSY <= '1';             -- Assert BUSY at start of frame
          
          when TRANSMIT_TAG =>                -- TRANSMIT_TAG state
            if(BIT_COUNTER = X"0F") then
              AC97_CONTROLLER <= TRANSMIT_SLOTS;
              INTERNAL_SYNC <= '0';           -- De-assert SYNC after TAG
            else
              AC97_CONTROLLER <= TRANSMIT_TAG;
            end if;
          



          when TRANSMIT_SLOTS =>              -- TRANSMIT_SLOTS state

            if(BIT_COUNTER >= 96) then
              INTERNAL_BUSY <= '0';           -- De-assert BUSY after 96 bits
            end if;
            if(BIT_COUNTER = X"FE") then
              AC97_CONTROLLER <= TRANSMIT_LAST_BIT_OF_FRAME;
              INTERNAL_SYNC <= '1';           -- Assert SYNC before last frame bit
            else
              AC97_CONTROLLER <= TRANSMIT_SLOTS;
            end if;

        end case; 

        BIT_COUNTER <= BIT_COUNTER + 1;
        
      end if;
    end if;
    
    if falling_edge(BIT_CLK) then
    
      if(RESET_IN = '0') then
        SHIFT_REGISTER <= (others => '0');
      else
        -- Only shift first 94 bits of input frame
        if((AC97_CONTROLLER /= TRANSMIT_LAST_BIT_OF_FRAME) and
           (BIT_COUNTER < 94)) then
              SHIFT_REGISTER <= SHIFT_REGISTER(1 to 93) & SDATA_IN;
        end if;
      end if;
      
    end if;
      
  end process;

  RESET_OUT <= RESET_IN;
  SYNC <= INTERNAL_SYNC;
  BUSY <= INTERNAL_BUSY;
  
  -- Only shift out 96 bits of relevant data
  SDATA_OUT <= FRAME_DATA(conv_integer(BIT_COUNTER(6 downto 0)))
               when (BIT_COUNTER < 94) else '0';

  -- Rename bits of SHIFT_REGISTER
  CODEC_READY <= SHIFT_REGISTER(0);
  VALID_STATUS_ADDRESS <= SHIFT_REGISTER(1);
  VALID_STATUS_DATA <= SHIFT_REGISTER(2);
  VALID_AUDIO_INPUT <= SHIFT_REGISTER(3) and SHIFT_REGISTER(4);
  STATUS_REGISTER_INDEX <= SHIFT_REGISTER(17 to 23);
  LEFT_AUDIO_REQUEST <= SHIFT_REGISTER(24);
  RIGHT_AUDIO_REQUEST <= SHIFT_REGISTER(25);
  STATUS_DATA <= SHIFT_REGISTER(36 to 51);
  LEFT_AUDIO_INPUT <= SHIFT_REGISTER(56 to 73);
  RIGHT_AUDIO_INPUT <= SHIFT_REGISTER(76 to 93);

end Behavioral;
