----------------------------------------------------------------------------------
-- BASIC_AC97_INTERFACE.VHD 
-- 
-- Steven Wasson
-- ECE 443 - Hardware Design with VHDL
-- Assignment 4
-- October 31, 2007
--
-- This VHDL module uses the BASIC_AC97_CORE to initialize the CODEC and
-- continuously transmit audio data to/from the CODEC
----------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity BASIC_AC97_INTERFACE is
  generic
  (
    SAMPLING_RATE                 : integer         :=  48000
  );
  port
  (
    -- Signals used to produce asynchronous reset to CODEC
    SYS_CLK                       : in  std_logic;  -- 100 MHz
    RESET_IN                      : in  std_logic;

    -- CODEC Signals
    BIT_CLK                       : in  std_logic;  -- 12.288 MHz
    RESET_OUT                     : out std_logic;
    SYNC                          : out std_logic;
    SDATA_OUT                     : out std_logic;
    SDATA_IN                      : in  std_logic;

    -- Interface Signals
    BUSY                          : out std_logic;
    READY_FOR_AUDIO_TRANSMISSION  : out std_logic;
    VALID_AUDIO_INPUT             : out std_logic;
    LEFT_AUDIO_INPUT              : out std_logic_vector(17 downto 0);
    RIGHT_AUDIO_INPUT             : out std_logic_vector(17 downto 0);
    VALID_AUDIO_OUTPUT            : in  std_logic;
    LEFT_AUDIO_OUTPUT             : in  std_logic_vector(17 downto 0);
    RIGHT_AUDIO_OUTPUT            : in  std_logic_vector(17 downto 0)    
  );
end BASIC_AC97_INTERFACE;

architecture Behavioral of BASIC_AC97_INTERFACE is

-- Define the BASIC_AC97_CORE component
component BASIC_AC97_CORE
  port
  (
    -- CODEC Signals
    BIT_CLK                         : in  std_logic;
    RESET_IN                        : in  std_logic;
    RESET_OUT                       : out std_logic;
    SYNC                            : out std_logic;
    SDATA_OUT                       : out std_logic;
    SDATA_IN                        : in  std_logic;
    BUSY                            : out std_logic;
    -- Input Interface Signals
    VALID_CONTROL_REGISTER_ADDRESS  : in  std_logic;
    VALID_CONTROL_REGISTER_DATA     : in  std_logic;
    VALID_AUDIO_OUTPUT              : in  std_logic;
    CONTROL_REGISTER_ADDRESS        : in  std_logic_vector(7 downto 0);
    CONTROL_REGISTER_DATA           : in  std_logic_vector(15 downto 0);
    LEFT_AUDIO_OUTPUT               : in  std_logic_vector(17 downto 0);
    RIGHT_AUDIO_OUTPUT              : in  std_logic_vector(17 downto 0);
    
    -- Output Interface Signals
    CODEC_READY                     : out std_logic;
    VALID_STATUS_ADDRESS            : out std_logic;
    VALID_STATUS_DATA               : out std_logic;
    VALID_AUDIO_INPUT               : out std_logic;
    STATUS_REGISTER_INDEX           : out std_logic_vector(6 downto 0);
    LEFT_AUDIO_REQUEST              : out std_logic;
    RIGHT_AUDIO_REQUEST             : out std_logic;
    STATUS_DATA                     : out std_logic_vector(15 downto 0);
    LEFT_AUDIO_INPUT                : out std_logic_vector(17 downto 0);
    RIGHT_AUDIO_INPUT               : out std_logic_vector(17 downto 0)
  );
end component;

-- Reset Generating Signals
signal    RESET_COUNTER                   : std_logic_vector(7 downto 0);
signal    AC97_RESET_IN                   : std_logic;

-- CODEC Interface signal
signal    INTERNAL_BUSY                   : std_logic;

-- Input fields from CODEC
signal    VALID_CONTROL_REGISTER_ADDRESS  : std_logic;
signal    VALID_CONTROL_REGISTER_DATA     : std_logic;
signal    CONTROL_REGISTER_ADDRESS        : std_logic_vector(7 downto 0);
signal    CONTROL_REGISTER_DATA           : std_logic_vector(15 downto 0);
signal    CODEC_READY                     : std_logic;
signal    VALID_STATUS_ADDRESS            : std_logic;
signal    VALID_STATUS_DATA               : std_logic;
signal    STATUS_REGISTER_INDEX           : std_logic_vector(6 downto 0);
signal    LEFT_AUDIO_REQUEST              : std_logic;
signal    RIGHT_AUDIO_REQUEST             : std_logic;
signal    STATUS_DATA                     : std_logic_vector(15 downto 0);

constant  SR  : std_logic_vector(15 downto 0) := conv_std_logic_vector(SAMPLING_RATE, 16);

-- Finite State Machine for controlling activity to CODEC
type    BASIC_AC97_AUDIO_OUT_SM is
        (WAIT_FOR_CODEC_READY, SET_MASTER_VOLUME, SET_HEAD_PHONE_VOLUME,
         SET_PCM_VOLUME, SET_RECORD_SELECT, SET_MIC_VOLUME, SET_RECORD_GAIN,
         SET_GENERAL_PURPOSE, SET_VRA, SET_DAC_RATE, SET_ADC_RATE,
         WAIT_FOR_BUSY_TOGO_LOW, WAIT_FOR_AUDIO_REQUEST, TRANSMIT_AUDIO);
signal  BASIC_AC97_AUDIO_OUT_CURRENT_TASK : BASIC_AC97_AUDIO_OUT_SM;
signal  BASIC_AC97_AUDIO_OUT_NEXT_TASK    : BASIC_AC97_AUDIO_OUT_SM;

begin





  BASIC_AC97_CORE_INSTANCE: BASIC_AC97_CORE
  port map
  (
    BIT_CLK                         => BIT_CLK,
    RESET_IN                        => AC97_RESET_IN,
    RESET_OUT                       => RESET_OUT,
    SYNC                            => SYNC,
    SDATA_OUT                       => SDATA_OUT,
    SDATA_IN                        => SDATA_IN,
    BUSY                            => INTERNAL_BUSY,

    VALID_CONTROL_REGISTER_ADDRESS  => VALID_CONTROL_REGISTER_ADDRESS,
    VALID_CONTROL_REGISTER_DATA     => VALID_CONTROL_REGISTER_DATA,
    VALID_AUDIO_OUTPUT              => VALID_AUDIO_OUTPUT,
    CONTROL_REGISTER_ADDRESS        => CONTROL_REGISTER_ADDRESS,
    CONTROL_REGISTER_DATA           => CONTROL_REGISTER_DATA,
    LEFT_AUDIO_OUTPUT               => LEFT_AUDIO_OUTPUT,
    RIGHT_AUDIO_OUTPUT              => RIGHT_AUDIO_OUTPUT,
    
    CODEC_READY                     => CODEC_READY,
    VALID_STATUS_ADDRESS            => VALID_STATUS_ADDRESS,
    VALID_STATUS_DATA               => VALID_STATUS_DATA,
    VALID_AUDIO_INPUT               => VALID_AUDIO_INPUT,
    STATUS_REGISTER_INDEX           => STATUS_REGISTER_INDEX,
    LEFT_AUDIO_REQUEST              => LEFT_AUDIO_REQUEST,
    RIGHT_AUDIO_REQUEST             => RIGHT_AUDIO_REQUEST,
    STATUS_DATA                     => STATUS_DATA,
    LEFT_AUDIO_INPUT                => LEFT_AUDIO_INPUT,
    RIGHT_AUDIO_INPUT               => RIGHT_AUDIO_INPUT
  );
  
  -- Generate a reset signal to the CODEC that lasts for 1.28 us
  -- when the RESET_IN line is asserted low
  GENERATE_RESET: process(SYS_CLK) is
  begin
    if rising_edge(SYS_CLK) then
      if(RESET_IN = '0') then
        RESET_COUNTER <= (others => '0');
        AC97_RESET_IN <= '0';
      else
        RESET_COUNTER <= RESET_COUNTER + '1';
        if(RESET_COUNTER(7) = '1') then
          AC97_RESET_IN <= '1';
        end if;
      end if;
    end if;
  end process;    
  
  INTERFACE_TO_AC97: process(BIT_CLK) is
  begin
    if(AC97_RESET_IN = '0') then
      BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= WAIT_FOR_CODEC_READY;
      BASIC_AC97_AUDIO_OUT_NEXT_TASK <= WAIT_FOR_CODEC_READY;
    
    else
      if rising_edge(BIT_CLK) then
                   
        case BASIC_AC97_AUDIO_OUT_CURRENT_TASK is
          

          -- Wait until the CODEC is ready before anything else
          when WAIT_FOR_CODEC_READY =>
            if((INTERNAL_BUSY & CODEC_READY) = "01") then
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= SET_MASTER_VOLUME;
            else
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= WAIT_FOR_CODEC_READY;
            end if;
          
          -- Clear the Master Volume MUTE bit and set the volume to 0 dB
          when SET_MASTER_VOLUME =>
            if(INTERNAL_BUSY = '0') then
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= SET_MASTER_VOLUME; 
            else
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_LOW;
              BASIC_AC97_AUDIO_OUT_NEXT_TASK <= SET_HEAD_PHONE_VOLUME;
            end if;
          
          -- Clear the Head Phone MUTE bit and set the volume to 0 dB
          when SET_HEAD_PHONE_VOLUME =>
            if(INTERNAL_BUSY = '0') then
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= SET_HEAD_PHONE_VOLUME; 
            else
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_LOW;
              BASIC_AC97_AUDIO_OUT_NEXT_TASK <= SET_PCM_VOLUME;
            end if;

          -- Clear the PCM Volume MUTE bit and set the volume to 0 dB
          when SET_PCM_VOLUME =>
            if(INTERNAL_BUSY = '0') then
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= SET_PCM_VOLUME; 
            else
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_LOW;
              BASIC_AC97_AUDIO_OUT_NEXT_TASK <= SET_RECORD_SELECT;
            end if;
          
          -- Set the Record Select to Microphone
          when SET_RECORD_SELECT =>
            if(INTERNAL_BUSY = '0') then
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= SET_RECORD_SELECT; 
            else
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_LOW;
              BASIC_AC97_AUDIO_OUT_NEXT_TASK <= SET_RECORD_GAIN;
            end if;
          
          -- Clear the Record Gain MUTE bit and set the volume to 10.5 dB
          when SET_RECORD_GAIN =>
            if(INTERNAL_BUSY = '0') then
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= SET_RECORD_GAIN; 
            else
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_LOW;
              BASIC_AC97_AUDIO_OUT_NEXT_TASK <= SET_MIC_VOLUME;
            end if;
          







          -- Enable the 20 dB gain for the Microphone
          when SET_MIC_VOLUME =>
            if(INTERNAL_BUSY = '0') then
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= SET_MIC_VOLUME; 
            else
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_LOW;
              BASIC_AC97_AUDIO_OUT_NEXT_TASK <= SET_GENERAL_PURPOSE;
            end if;
          
          -- Set the 3D bypass bit
          when SET_GENERAL_PURPOSE =>
            if(INTERNAL_BUSY = '0') then
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= SET_GENERAL_PURPOSE; 
            else
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_LOW;
              
              -- Set variable audio rate if sampling rate not equal to 48000 Hz
              -- Otherwise, use the frame rate (48000 Hz)
              if(SR /= 48000) then
                BASIC_AC97_AUDIO_OUT_NEXT_TASK <= SET_VRA;
              else
                BASIC_AC97_AUDIO_OUT_NEXT_TASK <= WAIT_FOR_AUDIO_REQUEST;
              end if;
            end if;
          
          -- Set the VRA bit
          when SET_VRA =>
            if(INTERNAL_BUSY = '0') then
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= SET_VRA; 
            else
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_LOW;
              BASIC_AC97_AUDIO_OUT_NEXT_TASK <= SET_DAC_RATE;
            end if;
          
          -- Set the PCM DAC Rate
          when SET_DAC_RATE =>
            if(INTERNAL_BUSY = '0') then
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= SET_DAC_RATE; 
            else
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_LOW;
              BASIC_AC97_AUDIO_OUT_NEXT_TASK <= SET_ADC_RATE;
            end if;
          
          -- Set the PCM ADC Rate
          when SET_ADC_RATE =>
            if(INTERNAL_BUSY = '0') then
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= SET_ADC_RATE; 
            else
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_LOW;
              BASIC_AC97_AUDIO_OUT_NEXT_TASK <= WAIT_FOR_AUDIO_REQUEST;
            end if;
          
          -- This state waits while the BASIC_AC97_CORE is busy
          when WAIT_FOR_BUSY_TOGO_LOW =>
            if(INTERNAL_BUSY = '0') then
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= BASIC_AC97_AUDIO_OUT_NEXT_TASK;
            else
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_LOW;
            end if;
          
          -- Wait until it is OK to transmit audio data
          when WAIT_FOR_AUDIO_REQUEST =>
            if((INTERNAL_BUSY & LEFT_AUDIO_REQUEST & RIGHT_AUDIO_REQUEST) = "000") then
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= TRANSMIT_AUDIO;
            else
              BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= WAIT_FOR_AUDIO_REQUEST;
            end if;

          -- Transmit audio data to CODEC
          when TRANSMIT_AUDIO =>
            if(INTERNAL_BUSY = '0') then
              if((LEFT_AUDIO_REQUEST & RIGHT_AUDIO_REQUEST) = "00") then
                BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= TRANSMIT_AUDIO;
              else
                BASIC_AC97_AUDIO_OUT_CURRENT_TASK <= WAIT_FOR_AUDIO_REQUEST;
              end if;
            end if;
            
        end case;       

      end if;
    end if;
  end process;      

  BUSY <= INTERNAL_BUSY;

  -- Assign values to the BASIC_AC97_CORE control lines
  -- depending on what task is being executed
  VALID_CONTROL_REGISTER_ADDRESS <=
    '1' when ((BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_MASTER_VOLUME) or
              (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_HEAD_PHONE_VOLUME) or
              (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_PCM_VOLUME) or
              (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_RECORD_SELECT) or
              (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_RECORD_GAIN) or
              (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_MIC_VOLUME) or
              (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_GENERAL_PURPOSE) or
              (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_VRA) or
              (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_DAC_RATE) or
              (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_ADC_RATE)) else '0';

  VALID_CONTROL_REGISTER_DATA <= VALID_CONTROL_REGISTER_ADDRESS;

  CONTROL_REGISTER_ADDRESS <=
    X"02" when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_MASTER_VOLUME) else
    X"04" when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_HEAD_PHONE_VOLUME) else
    X"18" when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_PCM_VOLUME) else
    X"1A" when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_RECORD_SELECT) else
    X"1C" when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_RECORD_GAIN) else
    X"0E" when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_MIC_VOLUME) else
    X"20" when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_GENERAL_PURPOSE) else
    X"2A" when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_VRA) else
    X"2C" when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_DAC_RATE) else
    X"32" when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_ADC_RATE) else
    X"00";






  CONTROL_REGISTER_DATA <=
    X"0808" when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_PCM_VOLUME) else
    X"8048" when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_MIC_VOLUME) else
    X"0707" when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_RECORD_GAIN) else
    X"8000" when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_GENERAL_PURPOSE) else
    X"0001" when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_VRA) else
    SR      when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_DAC_RATE) else
    SR      when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = SET_ADC_RATE) else
    X"0000";
  
  READY_FOR_AUDIO_TRANSMISSION <=
    '1' when (BASIC_AC97_AUDIO_OUT_CURRENT_TASK = TRANSMIT_AUDIO) else
    '0';

end Behavioral;
