----------------------------------------------------------------------------------
-- SOUND_RECORDER_CONTROLLER.VHD 
-- 
-- Steven Wasson
-- ECE 443 - Hardware Design with VHDL
-- Assignments 5 & 6
-- December 3, 2007
--
-- This VHDL module controls the audio recording/playing processes.
----------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

-- This package defines data types that are used to
-- control the audio and terminal interface processes
use sound_recorder_package.all;

entity RECORD_PLAY_CONTROLLER is
  port
  (
    RESET                         : in  std_logic;
    CLOCK                         : in  std_logic;
    
    COMMAND                       : in  RECORD_PLAY_COMMAND_TYPE;

    BUSY                          : in  std_logic;
    READY_FOR_AUDIO_TRANSMISSION  : in  std_logic;
    VALID_AUDIO_INPUT             : in  std_logic;
    AUDIO_INPUT                   : in  std_logic_vector(17 downto 0);
    AUDIO_OUTPUT                  : out std_logic_vector(17 downto 0);
    
    RECORD_IN_PROGRESS            : out std_logic;
    PLAY_IN_PROGRESS              : out std_logic;
    END_OF_AUDIO                  : out std_logic;
    ACKNOWLEDGE_EOA               : in  std_logic
  );
end RECORD_PLAY_CONTROLLER;

architecture Behavioral of RECORD_PLAY_CONTROLLER is

-- Define AUDIO_MEMORY BLOCK
-- This stores audio data captured from the AC97 CODEC
component AUDIO_MEMORY
	port
  (
    addr  : in  std_logic_VECTOR(16 downto 0);
    clk   : in  std_logic;
    din   : in  std_logic_VECTOR(17 downto 0);
    dout  : out std_logic_VECTOR(17 downto 0);
    we    : in  std_logic
  );
end component;

-- Audio Memory Signals
signal    AUDIO_MEMORY_POINTER          : std_logic_vector(16 downto 0);
signal    LAST_AUDIO_SAMPLE_ADDRESS     : std_logic_vector(16 downto 0);
signal    LOOPBACK_AUDIO_OUTPUT         : std_logic_vector(17 downto 0);
signal    RECORDED_AUDIO_OUTPUT         : std_logic_vector(17 downto 0);
signal    INTERNAL_RECORD_IN_PROGRESS   : std_logic;
signal    INTERNAL_PLAY_IN_PROGRESS     : std_logic;
signal    INTERNAL_END_OF_AUDIO         : std_logic;
signal    RECORD_ENABLE                 : std_logic;

-- Finite State Machine for controlling recording and playback activity
type      RECORD_PLAY_SM is
          (WAIT_FOR_COMMAND, WAIT_FOR_BUSY_TOGO_HIGH, WAIT_FOR_BUSY_TOGO_LOW,
           CHECK_CODEC_STATUS, RECORD_PLAY_SAMPLE);
signal    RECORD_PLAY_CURRENT_TASK      : RECORD_PLAY_SM;

begin

  LOOPBACK_AUDIO_PROCESS: process(CLOCK) is
  begin
    if rising_edge(CLOCK) then
      if((BUSY = '0') and (VALID_AUDIO_INPUT = '1')) then
        LOOPBACK_AUDIO_OUTPUT <= AUDIO_INPUT;
      end if;
    end if;
  end process;

  AUDIO_MEMORY_INSTANCE:
  AUDIO_MEMORY
  port map
  (
    addr  => AUDIO_MEMORY_POINTER,
    clk   => CLOCK,
    din   => AUDIO_INPUT,
    dout  => RECORDED_AUDIO_OUTPUT,
    we    => RECORD_ENABLE
  );
  
  AUDIO_OUTPUT <= LOOPBACK_AUDIO_OUTPUT when (INTERNAL_RECORD_IN_PROGRESS = '1') else
                  RECORDED_AUDIO_OUTPUT when (INTERNAL_PLAY_IN_PROGRESS = '1') else
                  (others => '0');

  RECORD_AND_PLAY:
  process(RESET, CLOCK) is
  begin
    if(RESET = '0') then
      AUDIO_MEMORY_POINTER <= (others => '0');
      INTERNAL_END_OF_AUDIO <= '0';
      RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_COMMAND;
      INTERNAL_RECORD_IN_PROGRESS <= '0';
      INTERNAL_PLAY_IN_PROGRESS <= '0';
      RECORD_ENABLE <= '0';
    
    elsif rising_edge(CLOCK) then
      
      case RECORD_PLAY_CURRENT_TASK is
      
        -- Wait until the RECORD or PLAY command has been issued
        when WAIT_FOR_COMMAND =>
          if((COMMAND = RECORD_COMMAND) and (INTERNAL_END_OF_AUDIO = '0')) then
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_HIGH;
            AUDIO_MEMORY_POINTER <= (others => '0');
            INTERNAL_END_OF_AUDIO <= '0';
            INTERNAL_RECORD_IN_PROGRESS <= '1';
          elsif((COMMAND = PLAY_COMMAND) and (INTERNAL_END_OF_AUDIO = '0')) then
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_HIGH;
            AUDIO_MEMORY_POINTER <= (others => '0');
            INTERNAL_END_OF_AUDIO <= '0';
            INTERNAL_PLAY_IN_PROGRESS <= '1';
          else
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_COMMAND;
          end if;
          
        -- Wait until BUSY goes high
        when WAIT_FOR_BUSY_TOGO_HIGH =>
          if(BUSY = '1') then
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_LOW;
          elsif(COMMAND = STOP_COMMAND) then
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_COMMAND;
            if(INTERNAL_RECORD_IN_PROGRESS = '1') then
              LAST_AUDIO_SAMPLE_ADDRESS <= AUDIO_MEMORY_POINTER;
            end if;
            INTERNAL_RECORD_IN_PROGRESS <= '0';
            INTERNAL_PLAY_IN_PROGRESS <= '0';
          else
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_HIGH;
          end if;
        
        -- Wait until BUSY goes low
        when WAIT_FOR_BUSY_TOGO_LOW =>
          if(BUSY = '0') then
            RECORD_PLAY_CURRENT_TASK <= CHECK_CODEC_STATUS;
          elsif(COMMAND = STOP_COMMAND) then
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_COMMAND;
            if(INTERNAL_RECORD_IN_PROGRESS = '1') then
              LAST_AUDIO_SAMPLE_ADDRESS <= AUDIO_MEMORY_POINTER;
            end if;
            INTERNAL_RECORD_IN_PROGRESS <= '0';
            INTERNAL_PLAY_IN_PROGRESS <= '0';
          else
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_LOW;
          end if;
        
        -- Check to see if the CODEC is ready for data
        when CHECK_CODEC_STATUS =>
          if((INTERNAL_RECORD_IN_PROGRESS = '1') and (VALID_AUDIO_INPUT = '1')) then
            RECORD_PLAY_CURRENT_TASK <= RECORD_PLAY_SAMPLE;
            RECORD_ENABLE <= '1';
          elsif((INTERNAL_PLAY_IN_PROGRESS = '1') and
                (READY_FOR_AUDIO_TRANSMISSION = '1')) then
            RECORD_PLAY_CURRENT_TASK <= RECORD_PLAY_SAMPLE;
          else
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_HIGH;
          end if;
        
        -- Record/Play the current audio sample
        when RECORD_PLAY_SAMPLE =>
          if((INTERNAL_RECORD_IN_PROGRESS = '1') and 
             (AUDIO_MEMORY_POINTER = X"1FFFF")) then
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_COMMAND;
            LAST_AUDIO_SAMPLE_ADDRESS <= (others => '1');
            INTERNAL_RECORD_IN_PROGRESS <= '0';
            INTERNAL_END_OF_AUDIO <= '1';
          elsif((INTERNAL_PLAY_IN_PROGRESS = '1') and
                (AUDIO_MEMORY_POINTER = LAST_AUDIO_SAMPLE_ADDRESS)) then
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_COMMAND;
            INTERNAL_PLAY_IN_PROGRESS <= '0';
            INTERNAL_END_OF_AUDIO <= '1';
          else
            RECORD_PLAY_CURRENT_TASK <= WAIT_FOR_BUSY_TOGO_HIGH;
          end if;
          
          AUDIO_MEMORY_POINTER <= AUDIO_MEMORY_POINTER + '1';
          RECORD_ENABLE <= '0';
          
      end case;
      
      if((INTERNAL_END_OF_AUDIO = '1') and (ACKNOWLEDGE_EOA = '1')) then
        INTERNAL_END_OF_AUDIO <= '0';
      end if;
    
    end if;
    
  end process;

  RECORD_IN_PROGRESS <= INTERNAL_RECORD_IN_PROGRESS;
  PLAY_IN_PROGRESS <= INTERNAL_PLAY_IN_PROGRESS;
  END_OF_AUDIO <= INTERNAL_END_OF_AUDIO;

end Behavioral;
