----------------------------------------------------------------------------------
-- SOUND_RECORDER.VHD 
-- 
-- Steven Wasson
-- ECE 443 - Hardware Design with VHDL
-- Assignments 5 & 6
-- December 2, 2007
--
-- This VHDL module records and plays audio through the AC97 Audio CODEC on the
-- XUP FPGA Board. The BASIC_AC97_CORE module is used to interface to the CODEC.
-- Single Port BRAM is used to store audio data from the CODEC. The record and
-- play processes can be controlled by two buttons or through an interactive RS232
-- terminal interface. The terminal interface uses cartoon characters from Looney
-- Tunes to interact with the user. These cartoon characters are represented as
-- ASCII Art obtained from http://www.heartnsoul.com/ascii_art/
-- ascii_cartoon_characters.htm.
----------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

-- This package defines data types that are used to
-- control the audio and terminal interface processes
use sound_recorder_package.all;

entity SOUND_RECORDER is
  port
  (
    -- Signals used to produce asynchronous reset to CODEC
    SYS_CLK               : in  std_logic;  -- 100 MHz
    RESET_IN              : in  std_logic;
    
    -- Record and Play Switches and LEDs
    BUTTON                : in  std_logic_vector(0 to 1);
    RECORD_LED            : out std_logic;
    PLAY_LED              : out std_logic;

    -- CODEC Signals
    BIT_CLK               : in  std_logic;  -- 12.288 MHz
    RESET_OUT             : out std_logic;
    SYNC                  : out std_logic;
    SDATA_OUT             : out std_logic;
    SDATA_IN              : in  std_logic;
    
    -- RS232 Signals
    TXD         : out std_logic;
    RXD         : in  std_logic
  );
end SOUND_RECORDER;

architecture Behavioral of SOUND_RECORDER is

-- Define SWITCH_DEBOUNCER component
-- This component debounces the record and play buttons
component SWITCH_DEBOUNCER
  generic
  (
    WIDTH                   : integer;
    DELAY                   : integer
  );

  port
  (
    RESET                   : in std_logic;
    CLOCK                   : in std_logic;
    INPUT                   : in std_logic_vector((WIDTH-1) downto 0);
    OUTPUT                  : out std_logic_vector((WIDTH-1) downto 0);
    NEW_OUTPUT              : out std_logic;
    ACKNOWLEDGE             : in std_logic
  );
end component;

-- Define CONSTANT_CLOCK_DIVIDER component
-- This generates a ~1MHz clock for the RS232 system
component CONSTANT_CLOCK_DIVIDER
  generic
  (
    DIVIDER       : integer
  );
  port
  (
    CLOCK_IN      : in  std_logic;
    CLOCK_OUT     : out std_logic
  );
end component;

-- Define BASIC_RS232_INTERFACE component
-- This provides a basic RS232 interface to a terminal
component BASIC_RS232_INTERFACE
  generic
  (
    INPUT_CLOCK_FREQUENCY : integer;
    BIT_RATE              : integer
  );
  port
  (
    RESET       : in  std_logic;
    CLOCK       : in  std_logic;

    RTS         : in  std_logic;  -- Request to Send
    DATA_OUT    : in  std_logic_vector(7 downto 0);
    TX_BUSY     : out std_logic;

    RX_INT      : out std_logic;  -- Receiver Interrupt
    FRAME_ERROR : out std_logic;
    RX_ACK      : in  std_logic;
    DATA_IN     : out std_logic_vector(7 downto 0);

    TXD         : out std_logic;
    RXD         : in  std_logic
  );
end component BASIC_RS232_INTERFACE;

-- This componont coordinates all the activity of the sound recorder
component SOUND_RECORDER_CONTROLLER
  port
  (
    RESET                   : in  std_logic;
    CLOCK                   : in  std_logic;
    
    RX_INT                  : in  std_logic;
    FRAME_ERROR             : in  std_logic;
    RX_ACK                  : out std_logic;
    ASCII_DATA              : in  std_logic_vector(7 downto 0);
    
    RECORD_BUTTON           : in  std_logic;
    PLAY_BUTTON             : in  std_logic;
    NEW_BUTTON_STATE        : in  std_logic;
    
    RECORD_IN_PROGRESS      : in  std_logic;
    PLAY_IN_PROGRESS        : in  std_logic;
    END_OF_AUDIO            : in  std_logic;
    ACKNOWLEDGE_EOA         : out std_logic;
    
    TERMINAL_OUTPUT_COMMAND : out TERMINAL_OUTPUT_COMMAND_TYPE;
    RECORD_PLAY_COMMAND     : out RECORD_PLAY_COMMAND_TYPE
  );
end component SOUND_RECORDER_CONTROLLER;

-- This component controls the interactive terminal interface
component TERMINAL_OUTPUT_INTERFACE
  port
  (
    RESET       : in  std_logic;
    CLOCK       : in  std_logic;
    
    COMMAND     : in  TERMINAL_OUTPUT_COMMAND_TYPE;

    RTS         : out std_logic;
    ASCII_DATA  : out std_logic_vector(7 downto 0);
    TX_BUSY     : in  std_logic
  );
end component TERMINAL_OUTPUT_INTERFACE;

-- Define BASIC_AC97_INTERFACE component
-- This interfaces to the AC97 CODEC
component BASIC_AC97_INTERFACE
  generic
  (
    SAMPLING_RATE                 : integer
  );
  port
  (
    -- Signals used to produce asynchronous reset to CODEC
    SYS_CLK                       : in  std_logic;  -- 100 MHz
    RESET_IN                      : in  std_logic;

    -- CODEC Signals
    BIT_CLK                       : in  std_logic;  -- 12.288 MHz
    RESET_OUT                     : out std_logic;
    SYNC                          : out std_logic;
    SDATA_OUT                     : out std_logic;
    SDATA_IN                      : in  std_logic;

    -- Interface Signals
    BUSY                          : out std_logic;
    READY_FOR_AUDIO_TRANSMISSION  : out std_logic;
    VALID_AUDIO_INPUT             : out std_logic;
    LEFT_AUDIO_INPUT              : out std_logic_vector(17 downto 0);
    RIGHT_AUDIO_INPUT             : out std_logic_vector(17 downto 0);
    VALID_AUDIO_OUTPUT            : in  std_logic;
    LEFT_AUDIO_OUTPUT             : in  std_logic_vector(17 downto 0);
    RIGHT_AUDIO_OUTPUT            : in  std_logic_vector(17 downto 0)    
  );
end component BASIC_AC97_INTERFACE;

-- This component controls the audio recording/playing processes
component RECORD_PLAY_CONTROLLER
  port
  (
    RESET                         : in  std_logic;
    CLOCK                         : in  std_logic;
    
    COMMAND                       : in  RECORD_PLAY_COMMAND_TYPE;

    BUSY                          : in  std_logic;
    READY_FOR_AUDIO_TRANSMISSION  : in  std_logic;
    VALID_AUDIO_INPUT             : in  std_logic;
    AUDIO_INPUT                   : in  std_logic_vector(17 downto 0);
    AUDIO_OUTPUT                  : out std_logic_vector(17 downto 0);
    
    RECORD_IN_PROGRESS            : out std_logic;
    PLAY_IN_PROGRESS              : out std_logic;
    END_OF_AUDIO                  : out std_logic;
    ACKNOWLEDGE_EOA               : in  std_logic
  );
end component RECORD_PLAY_CONTROLLER;

-- Record/Play button signals
signal    DEBOUNCED_BUTTON              : std_logic_vector(0 to 1);
signal    NEW_BUTTON_STATE              : std_logic;

-- Basic RS232 Interface Signals
signal    RS232_CLOCK                   : std_logic;
signal    RTS                           : std_logic;
signal    ASCII_DATA_OUT                : std_logic_vector(7 downto 0);
signal    TX_BUSY                       : std_logic;
signal    RX_INT                        : std_logic;
signal    FRAME_ERROR                   : std_logic;
signal    RX_ACK                        : std_logic;
signal    ASCII_DATA_IN                 : std_logic_vector(7 downto 0);

-- Basic AC97 Interface Signals
signal    BUSY                          : std_logic;
signal    READY_FOR_AUDIO_TRANSMISSION  : std_logic;
signal    VALID_AUDIO_INPUT             : std_logic;
signal    MONO_AUDIO_INPUT              : std_logic_vector(17 downto 0);
signal    MONO_AUDIO_OUTPUT             : std_logic_vector(17 downto 0);

-- Audio Memory Signals
signal    RECORD_IN_PROGRESS            : std_logic;
signal    PLAY_IN_PROGRESS              : std_logic;
signal    END_OF_AUDIO                  : std_logic;
signal    ACKNOWLEDGE_EOA               : std_logic;

-- Signals to control the TERMINAL_OUTPUT andRECORD_PLAY state machines
signal    TERMINAL_OUTPUT_COMMAND : TERMINAL_OUTPUT_COMMAND_TYPE;
signal    RECORD_PLAY_COMMAND     : RECORD_PLAY_COMMAND_TYPE;

begin

  -- Debounce the record/play buttons
  DEBOUNCE_BUTTONS: SWITCH_DEBOUNCER
  generic map
  (
    WIDTH                   => 2,
    DELAY                   => 4
  )
  port map
  (
    RESET                   => '0',
    CLOCK                   => BIT_CLK,
    INPUT                   => BUTTON,
    OUTPUT                  => DEBOUNCED_BUTTON,
    NEW_OUTPUT              => NEW_BUTTON_STATE,
    ACKNOWLEDGE             => NEW_BUTTON_STATE
  );
  
  GENERATE_RS232_CLOCK:
  CONSTANT_CLOCK_DIVIDER
  generic map
  (
    DIVIDER       => 6
  )
  port map
  (
    CLOCK_IN      => BIT_CLK,     -- 12.288 MHz
    CLOCK_OUT     => RS232_CLOCK  -- 1.024  MHz
  );
  
  BASIC_RS232_INTERFACE_INSTANCE:
  BASIC_RS232_INTERFACE
  generic map
  (
    INPUT_CLOCK_FREQUENCY => 1024000,
    BIT_RATE              => 115200
  )
  port map
  (
    RESET       => RESET_IN,
    CLOCK       => RS232_CLOCK,

    RTS         => RTS,
    DATA_OUT    => ASCII_DATA_OUT,
    TX_BUSY     => TX_BUSY,

    RX_INT      => RX_INT,
    FRAME_ERROR => FRAME_ERROR,
    RX_ACK      => RX_ACK,
    DATA_IN     => ASCII_DATA_IN,

    TXD         => TXD,
    RXD         => RXD
  );
  
  SOUND_RECORDER_CONTROLLER_INSTANCE:
  SOUND_RECORDER_CONTROLLER
  port map
  (
    RESET                   => RESET_IN,
    CLOCK                   => BIT_CLK,
    
    RX_INT                  => RX_INT,
    FRAME_ERROR             => FRAME_ERROR,
    RX_ACK                  => RX_ACK,
    ASCII_DATA              => ASCII_DATA_IN,
    
    RECORD_BUTTON           => DEBOUNCED_BUTTON(0),
    PLAY_BUTTON             => DEBOUNCED_BUTTON(1),
    NEW_BUTTON_STATE        => NEW_BUTTON_STATE,
    
    RECORD_IN_PROGRESS      => RECORD_IN_PROGRESS,
    PLAY_IN_PROGRESS        => PLAY_IN_PROGRESS,
    END_OF_AUDIO            => END_OF_AUDIO,
    ACKNOWLEDGE_EOA         => ACKNOWLEDGE_EOA,
    
    TERMINAL_OUTPUT_COMMAND => TERMINAL_OUTPUT_COMMAND,
    RECORD_PLAY_COMMAND     => RECORD_PLAY_COMMAND
  );
  
  TERMINAL_OUTPUT_INTERFACE_INSTANCE:
  TERMINAL_OUTPUT_INTERFACE
  port map
  (
    RESET       =>  RESET_IN,
    CLOCK       =>  BIT_CLK,
    
    COMMAND     =>  TERMINAL_OUTPUT_COMMAND,

    RTS         =>  RTS,
    ASCII_DATA  =>  ASCII_DATA_OUT,
    TX_BUSY     =>  TX_BUSY
  );
  
  BASIC_AC97_INTERFACE_INSTANCE:
  BASIC_AC97_INTERFACE
  generic map
  (
    SAMPLING_RATE                 => 4000 -- Hz
  )
  port map
  (
    -- Signals used to produce asynchronous reset to CODEC
    SYS_CLK                       => SYS_CLK,
    RESET_IN                      => RESET_IN,

    -- CODEC Signals
    BIT_CLK                       => BIT_CLK,
    RESET_OUT                     => RESET_OUT,
    SYNC                          => SYNC,
    SDATA_OUT                     => SDATA_OUT,
    SDATA_IN                      => SDATA_IN,

    -- Interface Signals
    BUSY                          => BUSY,
    READY_FOR_AUDIO_TRANSMISSION  => READY_FOR_AUDIO_TRANSMISSION,----
    VALID_AUDIO_INPUT             => VALID_AUDIO_INPUT,--             }
    LEFT_AUDIO_INPUT              => MONO_AUDIO_INPUT,--              |
    RIGHT_AUDIO_INPUT             => open,--                          }
    VALID_AUDIO_OUTPUT            => READY_FOR_AUDIO_TRANSMISSION,----
    LEFT_AUDIO_OUTPUT             => MONO_AUDIO_OUTPUT,
    RIGHT_AUDIO_OUTPUT            => MONO_AUDIO_OUTPUT
  );
  
  RECORD_PLAY_CONTROLLER_INSTANCE:
  RECORD_PLAY_CONTROLLER
  port map
  (
    RESET                         => RESET_IN,
    CLOCK                         => BIT_CLK,
    
    COMMAND                       => RECORD_PLAY_COMMAND,

    BUSY                          => BUSY,
    READY_FOR_AUDIO_TRANSMISSION  => READY_FOR_AUDIO_TRANSMISSION,
    VALID_AUDIO_INPUT             => VALID_AUDIO_INPUT,
    AUDIO_INPUT                   => MONO_AUDIO_INPUT,
    AUDIO_OUTPUT                  => MONO_AUDIO_OUTPUT,
    
    RECORD_IN_PROGRESS            => RECORD_IN_PROGRESS,
    PLAY_IN_PROGRESS              => PLAY_IN_PROGRESS,
    END_OF_AUDIO                  => END_OF_AUDIO,
    ACKNOWLEDGE_EOA               => ACKNOWLEDGE_EOA
  );
  
  RECORD_LED <= not RECORD_IN_PROGRESS;
  PLAY_LED <= not PLAY_IN_PROGRESS;

end Behavioral;
