----------------------------------------------------------------------------------
-- SOUND_RECORDER_CONTROLLER.VHD 
-- 
-- Steven Wasson
-- ECE 443 - Hardware Design with VHDL
-- Assignments 5 & 6
-- December 3, 2007
--
-- This VHDL module coordinates all the activity of the sound recorder.
----------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

-- This package defines data types that are used to
-- control the audio and terminal interface processes
use sound_recorder_package.all;

entity SOUND_RECORDER_CONTROLLER is
  port
  (
    RESET                   : in  std_logic;
    CLOCK                   : in  std_logic;
    
    RX_INT                  : in  std_logic;
    FRAME_ERROR             : in  std_logic;
    RX_ACK                  : out std_logic;
    ASCII_DATA              : in  std_logic_vector(7 downto 0);
    
    RECORD_BUTTON           : in  std_logic;
    PLAY_BUTTON             : in  std_logic;
    NEW_BUTTON_STATE        : in  std_logic;
    
    RECORD_IN_PROGRESS      : in  std_logic;
    PLAY_IN_PROGRESS        : in  std_logic;
    END_OF_AUDIO            : in  std_logic;
    ACKNOWLEDGE_EOA         : out std_logic;
    
    TERMINAL_OUTPUT_COMMAND : out TERMINAL_OUTPUT_COMMAND_TYPE;
    RECORD_PLAY_COMMAND     : out RECORD_PLAY_COMMAND_TYPE
  );
end SOUND_RECORDER_CONTROLLER;

architecture Behavioral of SOUND_RECORDER_CONTROLLER is

signal    INTERNAL_TERMINAL_OUTPUT_COMMAND  :   TERMINAL_OUTPUT_COMMAND_TYPE
                                            :=  OUTPUT_BUGS_BUNNY_1;

-- Finite State Machine for controlling the whole system
type      SOUND_RECORDER_SM is 
          (WAIT_FOR_INPUT, ACKNOWLEDGE_TERMINAL_INPUT,
           WAIT_FOR_INACTIVE_BUTTONS, ACKNOWLEDGE_END_OF_AUDIO, EXIT_PROGRAM);
signal    SOUND_RECORDER_CURRENT_TASK   : SOUND_RECORDER_SM := WAIT_FOR_INPUT;

begin

  SOUND_RECORDER_PROCESS:
  process(CLOCK) is
  begin
    if(RESET = '0') then
      SOUND_RECORDER_CURRENT_TASK <= WAIT_FOR_INPUT;
      INTERNAL_TERMINAL_OUTPUT_COMMAND <= IDLE;
      RECORD_PLAY_COMMAND <= STOP_COMMAND;

    elsif rising_edge(CLOCK) then
      
      case SOUND_RECORDER_CURRENT_TASK is
        
        when WAIT_FOR_INPUT =>
          if(RX_INT = '1') then
            SOUND_RECORDER_CURRENT_TASK <= ACKNOWLEDGE_TERMINAL_INPUT;
            if(FRAME_ERROR = '0') then
              case ASCII_DATA is
                when X"48" => -- H
                  INTERNAL_TERMINAL_OUTPUT_COMMAND <= OUTPUT_BUGS_BUNNY_1;
                when X"52" => -- R (Record)
                  INTERNAL_TERMINAL_OUTPUT_COMMAND <= OUTPUT_PEPE_LEPEW;
                  RECORD_PLAY_COMMAND <= RECORD_COMMAND;
                when X"50" => -- P (Play)
                  INTERNAL_TERMINAL_OUTPUT_COMMAND <= OUTPUT_TWEETY_BIRD;
                  RECORD_PLAY_COMMAND <= PLAY_COMMAND;
                when X"53" => -- S (Stop)
                  INTERNAL_TERMINAL_OUTPUT_COMMAND <= OUTPUT_BUGS_BUNNY_2;
                  RECORD_PLAY_COMMAND <= STOP_COMMAND;
                when X"58" => -- X (Exit)
                  INTERNAL_TERMINAL_OUTPUT_COMMAND <= OUTPUT_TASMANIAN_DEVIL;
                when others =>
                  INTERNAL_TERMINAL_OUTPUT_COMMAND <= OUTPUT_DAFFY_DUCK;
              end case;
            end if;
          elsif(NEW_BUTTON_STATE = '1') then
            SOUND_RECORDER_CURRENT_TASK <= WAIT_FOR_INACTIVE_BUTTONS;
            if(RECORD_BUTTON = '0') then  -- Record button has been pushed
              if((RECORD_IN_PROGRESS = '0') and (PLAY_IN_PROGRESS = '0')) then
                INTERNAL_TERMINAL_OUTPUT_COMMAND <= OUTPUT_PEPE_LEPEW;
                RECORD_PLAY_COMMAND <= RECORD_COMMAND;
              else
                INTERNAL_TERMINAL_OUTPUT_COMMAND <= OUTPUT_BUGS_BUNNY_2;
                RECORD_PLAY_COMMAND <= STOP_COMMAND;
              end if;
            elsif(PLAY_BUTTON = '0') then -- Play button has been pushed
              if((PLAY_IN_PROGRESS = '0') and (RECORD_IN_PROGRESS = '0')) then
                INTERNAL_TERMINAL_OUTPUT_COMMAND <= OUTPUT_TWEETY_BIRD;
                RECORD_PLAY_COMMAND <= PLAY_COMMAND;
              else
                INTERNAL_TERMINAL_OUTPUT_COMMAND <= OUTPUT_BUGS_BUNNY_2;
                RECORD_PLAY_COMMAND <= STOP_COMMAND;
              end if;
            else
              INTERNAL_TERMINAL_OUTPUT_COMMAND <= OUTPUT_DAFFY_DUCK;
            end if;
          elsif(END_OF_AUDIO = '1') then  -- Recorded/Played all audio
            SOUND_RECORDER_CURRENT_TASK <= ACKNOWLEDGE_END_OF_AUDIO;
            INTERNAL_TERMINAL_OUTPUT_COMMAND <= OUTPUT_BUGS_BUNNY_2;
            RECORD_PLAY_COMMAND <= STOP_COMMAND;
          else
            SOUND_RECORDER_CURRENT_TASK <= WAIT_FOR_INPUT;
            INTERNAL_TERMINAL_OUTPUT_COMMAND <= IDLE;
          end if;
          
        when ACKNOWLEDGE_TERMINAL_INPUT =>

          INTERNAL_TERMINAL_OUTPUT_COMMAND <= IDLE;
          if(RX_INT = '0') then
            if(INTERNAL_TERMINAL_OUTPUT_COMMAND /= OUTPUT_TASMANIAN_DEVIL) then
              SOUND_RECORDER_CURRENT_TASK <= WAIT_FOR_INPUT;
            else
              SOUND_RECORDER_CURRENT_TASK <= EXIT_PROGRAM;
            end if;
          else
            SOUND_RECORDER_CURRENT_TASK <= ACKNOWLEDGE_TERMINAL_INPUT;
          end if;
          
        when WAIT_FOR_INACTIVE_BUTTONS =>

          INTERNAL_TERMINAL_OUTPUT_COMMAND <= IDLE;
          if((RECORD_BUTTON = '1') and (PLAY_BUTTON = '1')) then
            SOUND_RECORDER_CURRENT_TASK <= WAIT_FOR_INPUT;
          else
            SOUND_RECORDER_CURRENT_TASK <= WAIT_FOR_INACTIVE_BUTTONS;
          end if;
        
        when ACKNOWLEDGE_END_OF_AUDIO =>

          INTERNAL_TERMINAL_OUTPUT_COMMAND <= IDLE;
          if(END_OF_AUDIO = '0') then
            SOUND_RECORDER_CURRENT_TASK <= WAIT_FOR_INPUT;
          else
            SOUND_RECORDER_CURRENT_TASK <= ACKNOWLEDGE_END_OF_AUDIO;
          end if;
        
        when EXIT_PROGRAM =>
          SOUND_RECORDER_CURRENT_TASK <= EXIT_PROGRAM;
          INTERNAL_TERMINAL_OUTPUT_COMMAND <= IDLE;
          RECORD_PLAY_COMMAND <= STOP_COMMAND;
          
      end case;

    end if;
  end process;
  
  TERMINAL_OUTPUT_COMMAND <= INTERNAL_TERMINAL_OUTPUT_COMMAND;
  
  RX_ACK <= '1'
    when (SOUND_RECORDER_CURRENT_TASK = ACKNOWLEDGE_TERMINAL_INPUT) else '0';

  ACKNOWLEDGE_EOA <= '1'
    when (SOUND_RECORDER_CURRENT_TASK = ACKNOWLEDGE_END_OF_AUDIO) else '0';

end Behavioral;
