----------------------------------------------------------------------------------
-- TERMINAL_OUTPUT_INTERFACE.VHD 
-- 
-- Steven Wasson
-- ECE 443 - Hardware Design with VHDL
-- Assignments 5 & 6
-- December 3, 2007
--
-- This VHDL module controls the interactive terminal interface.
----------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

-- This package defines data types that are used to
-- control the audio and terminal interface processes
use sound_recorder_package.all;

entity TERMINAL_OUTPUT_INTERFACE is
  port
  (
    RESET       : in  std_logic;
    CLOCK       : in  std_logic;
    
    COMMAND     : in  TERMINAL_OUTPUT_COMMAND_TYPE;

    RTS         : out std_logic;
    ASCII_DATA  : out std_logic_vector(7 downto 0);
    TX_BUSY     : in  std_logic
  );
end TERMINAL_OUTPUT_INTERFACE;

architecture Behavioral of TERMINAL_OUTPUT_INTERFACE is

-- Define ASCII_ROM component
-- This stores all the terminal interface data
component ASCII_ROM
	port
  (
    clka:   in  std_logic;
    addra:  in  std_logic_VECTOR(13 downto 0);
    douta:  out std_logic_VECTOR(6 downto 0)
  );
end component;

-- Finite State Machine for outputting ASCII art to terminal interface
type      TERMINAL_OUTPUT_CONTROLLER_SM is 
          (WAIT_FOR_COMMAND, WAIT_FOR_TX_BUSY_TOGO_HIGH, WAIT_FOR_TX_BUSY_TOGO_LOW,
           TRANSMIT_CHARACTERS, INCREMENT_ADDRESS);
signal    TERMINAL_OUTPUT_CURRENT_TASK  : TERMINAL_OUTPUT_CONTROLLER_SM;

-- ASCII Memory Signals
signal    ASCII_MEMORY_POINTER          : std_logic_vector(13 downto 0);
signal    LAST_ASCII_MEMORY_ADDRESS     : std_logic_vector(13 downto 0);
constant  BUGS_BUNNY_1    : std_logic_vector(13 downto 0) := "00000000000000";
constant  PEPE_LEPEW      : std_logic_vector(13 downto 0) := "00100101110010";
constant  TWEETY_BIRD     : std_logic_vector(13 downto 0) := "01000000101000";
constant  BUGS_BUNNY_2    : std_logic_vector(13 downto 0) := "01100010010110";
constant  TASMANIAN_DEVIL : std_logic_vector(13 downto 0) := "01111011100011";
constant  DAFFY_DUCK      : std_logic_vector(13 downto 0) := "10001101110000";
constant  ASCII_ROM_SIZE  : std_logic_vector(13 downto 0) := "10100100000100";

begin

  ASCII_ART_INTERFACE:
  ASCII_ROM
	port map
  (
    clka  => CLOCK,
    addra => ASCII_MEMORY_POINTER,
    douta => ASCII_DATA(6 downto 0)
  );
  
  ASCII_DATA(7) <= '0';
  
  TERMINAL_OUTPUT_INTERFACE_CONTROLLER:
  process(RESET, CLOCK) is
  begin
    if(RESET = '0') then
      RTS <= '0';
      ASCII_MEMORY_POINTER <= BUGS_BUNNY_1;
      LAST_ASCII_MEMORY_ADDRESS <= PEPE_LEPEW;
      TERMINAL_OUTPUT_CURRENT_TASK <= TRANSMIT_CHARACTERS;

    elsif rising_edge(CLOCK) then
    
      case TERMINAL_OUTPUT_CURRENT_TASK is
      
        when WAIT_FOR_COMMAND =>
          
          if(COMMAND /= IDLE) then
            
            TERMINAL_OUTPUT_CURRENT_TASK <= TRANSMIT_CHARACTERS;
              
            case COMMAND is
              when OUTPUT_BUGS_BUNNY_1 =>
                ASCII_MEMORY_POINTER <= BUGS_BUNNY_1;
                LAST_ASCII_MEMORY_ADDRESS <= PEPE_LEPEW;
              when OUTPUT_PEPE_LEPEW =>
                ASCII_MEMORY_POINTER <= PEPE_LEPEW;
                LAST_ASCII_MEMORY_ADDRESS <= TWEETY_BIRD;
              when OUTPUT_TWEETY_BIRD =>
                ASCII_MEMORY_POINTER <= TWEETY_BIRD;
                LAST_ASCII_MEMORY_ADDRESS <= BUGS_BUNNY_2;
              when OUTPUT_BUGS_BUNNY_2 =>
                ASCII_MEMORY_POINTER <= BUGS_BUNNY_2;
                LAST_ASCII_MEMORY_ADDRESS <= TASMANIAN_DEVIL;
              when OUTPUT_TASMANIAN_DEVIL =>
                ASCII_MEMORY_POINTER <= TASMANIAN_DEVIL;
                LAST_ASCII_MEMORY_ADDRESS <= DAFFY_DUCK;
              when OUTPUT_DAFFY_DUCK =>
                ASCII_MEMORY_POINTER <= DAFFY_DUCK;
                LAST_ASCII_MEMORY_ADDRESS <= ASCII_ROM_SIZE;
              when others =>
                ASCII_MEMORY_POINTER <= BUGS_BUNNY_1;
                LAST_ASCII_MEMORY_ADDRESS <= PEPE_LEPEW;
            end case;
            
          else
            
            TERMINAL_OUTPUT_CURRENT_TASK <= WAIT_FOR_COMMAND;
              
          end if;

        when WAIT_FOR_TX_BUSY_TOGO_HIGH =>
          if(TX_BUSY = '1') then
            RTS <= '0';
            TERMINAL_OUTPUT_CURRENT_TASK <= WAIT_FOR_TX_BUSY_TOGO_LOW;
          else
            TERMINAL_OUTPUT_CURRENT_TASK <= WAIT_FOR_TX_BUSY_TOGO_HIGH;
          end if;
        
        when WAIT_FOR_TX_BUSY_TOGO_LOW =>
          if(TX_BUSY = '0') then
            TERMINAL_OUTPUT_CURRENT_TASK <= INCREMENT_ADDRESS;
          else
            TERMINAL_OUTPUT_CURRENT_TASK <= WAIT_FOR_TX_BUSY_TOGO_LOW;
          end if;

        when TRANSMIT_CHARACTERS =>
          if(ASCII_MEMORY_POINTER /= LAST_ASCII_MEMORY_ADDRESS) then
            TERMINAL_OUTPUT_CURRENT_TASK <= WAIT_FOR_TX_BUSY_TOGO_HIGH;
            RTS <= '1';
          else
            TERMINAL_OUTPUT_CURRENT_TASK <= WAIT_FOR_COMMAND;
          end if;
        
        when INCREMENT_ADDRESS =>
          TERMINAL_OUTPUT_CURRENT_TASK <= TRANSMIT_CHARACTERS;
          ASCII_MEMORY_POINTER <= ASCII_MEMORY_POINTER + '1';

      end case;

    end if;
  end process;
  
end Behavioral;

